# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.

""" Helper functions and classes to deal with files and dirs and stuff

Featuring::

* the cd context manager pwd(); with cd(other_dir): pwd();   pwd();
  will print current working dir, then other_dir, then first dir again
* a wrapper around "df" to determine size and usage of file systems

.. todo:: change get_filesystem_fill_states to not fork

.. codeauthor:: Intra2net
"""

from __future__ import print_function

from contextlib import contextmanager
import os
from warnings import warn
from math import floor, ceil

from call_helpers import call_and_capture
from iter_helpers import pairwise


@contextmanager
def cd(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    Taken from comment for python recipe by Greg Warner at
    http://code.activestate.com/recipes/576620-changedirectory-context-manager/
    (MIT license)
    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


DF_CMD = ['/usr/bin/df', '--no-sync', '--portability']
DF_SIZE_UNIT = 1024

class FilesystemFillState:
    """ representation of 1 line of the 'df' command

    has fields filesystem, size, used, available, capacity, mount_point

    Note that only apprixomately capacity == used/size
     and that only approximately used + available == size
     and that all sizes are in bytes
    """

    def __init__(self):
        self.name = None
        self.size = None
        self.used = None
        self.available = None
        self.capacity = None
        self.mount_point = None

    def __str__(self):
        return '[Filesystem {0} mounted at {1}: {2}% used]' \
               .format(self.name, self.mount_point, self.capacity)

def get_filesystem_fill_states():
    """ get fill state on all filesystems

    parses the output of cmd 'df', returns list of
    :py:class:`FilesystemFillState`

    ..todo:: replace call to df (which forks) by call to statvfs combined with
             some way to find all mounted partitions (parse /proc/mounts
             whenever it changes)
    """

    # call
    code, out, err = call_and_capture(DF_CMD)

    # warn if unexpected outcome
    if code != 0:
        warn('df returned non-zero exit code {0}!'.format(code))
    if err:
        for line in err:
            warn('df produced output to stderr: "{0}"'.format(line))

    # find columns in output that are just spaces
    min_len = min(len(line) for line in out)
    separator_cols = [idx for idx in range(min_len) if \
                      all(line[idx] == ' ' for line in out)]
    checked_cols = [separator_cols[0], ]
    for prev_col, col in pairwise(separator_cols):
        if col != prev_col+1:
            checked_cols.append(col)
    separator_cols = checked_cols

    # check columns and their header
    if len(separator_cols) != 5:
        raise ValueError('unexpected number of separator columns: {0}'
                         .format(separator_cols))  # must eliminate neighbours?

    title_line = out[0]
    title = title_line[ : separator_cols[0]].strip()
    if title != 'Filesystem':
        warn('Unexpected column title: "{0}" != "Filesystem"!'
             .format(title))
    title = title_line[separator_cols[0] : separator_cols[1]].strip()
    if title != '1024-blocks':
        warn('Unexpected column title: "{0}" != "1024-blocks"!'
             .format(title))
    title = title_line[separator_cols[1] : separator_cols[2]].strip()
    if title != 'Used':
        warn('Unexpected column title: "{0}" != "Used"!'
             .format(title))
    title = title_line[separator_cols[2] : separator_cols[3]].strip()
    if title != 'Available':
        warn('Unexpected column title: "{0}" != "Available"!'
             .format(title))
    title = title_line[separator_cols[3] : separator_cols[4]].strip()
    if title != 'Capacity':
        warn('Unexpected column title: "{0}" != "Capacity"!'
             .format(title))
    title = title_line[separator_cols[4] : ].strip()
    if title != 'Mounted on':
        warn('Unexpected column title: "{0}" != "Mounted on"!'
             .format(title))

    # create result
    result = []
    for line in out[1:]:
        stats = FilesystemFillState()
        stats.name = line[ : separator_cols[0]].strip()
        stats.size = int(line[separator_cols[0] : separator_cols[1]].strip()) \
                     * DF_SIZE_UNIT
        stats.used = int(line[separator_cols[1] : separator_cols[2]].strip()) \
                     * DF_SIZE_UNIT
        stats.available = int(line[separator_cols[2] : separator_cols[3]]\
                              .strip()) * DF_SIZE_UNIT
        stats.capacity = int(line[separator_cols[3] : separator_cols[4]]\
                             .strip()[:-1])
        stats.mount_point = line[separator_cols[4] : ].strip()

        # more checks: does match capacity
        capacity = 100. * stats.used / stats.size
        if abs(capacity - stats.capacity) > 5.:
            warn('capacities for {0} deviate more than 5%: '
                 '{1} != {2:.2f}(={3}/{4})'.format(
                 stats.name, stats.capacity, capacity, stats.used, stats.size))

        size = stats.used + stats.available
        if float(abs(stats.size - size)) / float(max(stats.size, size)) > 0.1:
            warn('size for {0} differs by more than 10% from used+available!'
                 .format(stats.name))

        result.append(stats)

    return result
